### genetische Selbstorganisation Variante 4.py
from random import *
# Es werden in einem Quadrat zufaellig Punkte erzeugt. 
# Dann werden sie einem genetischen Prozess unterworfen,
# bei dem als Bewertungsfunktion die Anzahl der Punkte
# verwendet wird, die untereinander ein Abstandsquadrat<eps
# haben.
# Also: Kein Kreismerkmal zu erwarten!

import tkinter.filedialog as fileDialog
fg,bg,punktFarbe='black','white','blue'
# fuer schwarzen Hintergrund umstellen:
fg,bg,punktFarbe='white','black','yellow'
        
### ===== zufallsgeneratorSetzen ===================================
# Setzt den Zufallsgenerator je nach Parameter mit einer "zufaelligen"
# Zahl oder der Zahl 123456789.
def zufallsgeneratorSetzen(zeit):
    if zeit: seed()
    else: seed(zeit)

### ===== erzeugePunkte ============================================
# Erzeugt eine Liste von n zufaelligen Punkten im Quadrat der Breite
# 'breite'. Alle Koordinaten sind 0 <= x,y < breite
def erzeugePunkte(n, breite, zeit):
    zufallsgeneratorSetzen(zeit)
    punkte=[]
    for i in range(n):
        x,y=randint(0,breite-1),randint(0,breite-1)
        punkte.append((x,y))
    return punkte

### ===== abstandUrsprung ==========================================
# Berechnet jeweils zu jedem Punkt den Abstand vom Ursprung (allerdings
# wegen der Rechenzeit das Quadrat) und gibt Punkte und Abstaende
# als Dictionary zurueck (was hier eigentlich nicht gebraucht wird).
def abstandUrsprung(punkte):
    def abstand(x,y):
        return x*x + y*y
    d={}
    for punkt in punkte:
        d.update({punkt:abstand(punkt[0],punkt[1])})
    return d

### ===== bewerteAbstaende =========================================
def bewerteAbstaende(punktDict, eps):
    # gibt die Anzahl von Punkten in der Punktliste zurueck,
    # die untereinander ein Abstandsquadrat<eps haben;
    # hier wird das Dictionary also eigentlich nicht gebraucht
    anzahl=0
    for punkt1 in punktDict:
        x1,y1=punkt1
        for punkt2 in punktDict:
            x2,y2=punkt2
            if (x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)>eps: #alternativ:
##            if (x1-x2)*(x1-x2)+(y1-y2)*(y1-y2)<eps: #alternativ:
                anzahl+=1
    return anzahl

### ===== heize =====================================================
# nimmt einen Punkt und veraendert die Koordinaten mit zufaelligen,
# aber von dem Parameter Temperatur abhaengigen Veraenderungen.
def heize(punkt, temperatur, alleWieviel):
    if randint(0,alleWieviel-1)==0:
        punkt=heizeKoordinate(punkt[0],temperatur),\
               heizeKoordinate(punkt[1],temperatur)
    return punkt

### ===== heizeKoordinate =========================================
def heizeKoordinate(koordinate, temperatur):
    erg=abs(random()*temperatur + koordinate)
    erg-=temperatur/2
    return round(erg)
    
### ===== einSchritt ==============================================
# Nach dem Heizen werden die urspruengliche Loesung und die geheizte
# verglichen und liefert das Dictionary zurueck, welche die
# groessere Bewertungszahl hat.
def einSchritt(punktDict, temperatur, eps, alleWieviel):
    neuDict=heizeDict(punktDict, temperatur, alleWieviel) # geheizte Kopie
    if bewerteAbstaende(neuDict, eps)<bewerteAbstaende(punktDict, eps):
        return neuDict
    return punktDict

### ===== punkteAusDict ===========================================
# Erstellt eine geheizte Kopie zum Punkt-Dictionary 
def heizeDict(punktDict, temperatur, alleWieviel):
    neuePunkte=[] # Kopie zum Heizen
    for punkt in holePunkte(punktDict):
        neuePunkte.append(heize(punkt, temperatur, alleWieviel))
    neuDict=abstandUrsprung(neuePunkte)
    return neuDict

def holePunkte(punktDict):
    neuePunkte=[]
    for punkt in punktDict:
        neuePunkte.append(punkt)
    return neuePunkte

### ===== schritte ============================================
# Fuehrt das Verfahren fuer n Punkte m Schritte lang durch.
def schritte(n, m, temperatur, eps, alleWieviel, zeit, breite):
    punktDict=abstandUrsprung(erzeugePunkte(n, breite, zeit))
    daten=[]
    for i in range(m):
        if i % alleWieviel==0:
            daten.append(holePunkte(punktDict)) # werden gezeichnet
            print('Schritt:',i)
        punktDict=einSchritt(punktDict, temperatur, eps, alleWieviel)
    return daten        
  
### ----- Grafikabschnitt --------------------------------------
from tkinter import *
import _thread
import time

class grafikApp(Frame):

    def __init__(self, master=None, breite=400):
        '''breite (= hoehe) definert die Abmessungen der Zeichenflaeche
        Vorgabe ist 400 x 400'''
        Frame.__init__(self, master)
        self.breite = breite
        self.nPunkte=200
        self.mSchritte=2000
        self.temperatur=100
        self.eps=2000
        self.alleWieviel=50
        self.daten=schritte(self.nPunkte,\
                            self.mSchritte,\
                            self.temperatur,\
                            self.eps, self.alleWieviel,
                            False, self.breite)
        # n, m, temperatur, eps, alleWieviel, zeit, breite
        self.grid()
        self.createWidgets()
        self.zeichne()

    def createWidgets(self):
        self.zeichenfenster = Canvas(self,height=self.breite,width=self.breite,bg=bg)
        self.zeichenfenster.grid(row=0, column=0)
        self.zeichenfenster.bind("<Double-Button-1>", self.onBild)
        self.beenden_Button = Button(self, text= "Beenden", command=self.quit, width=15)
        self.beenden_Button.grid(row=1, column=0, pady=3)

    def onBild(self, evt=None):
        dateiname=fileDialog.asksaveasfilename(\
            parent=self,
            initialfile="Bild.ps",
            defaultextension='ps',
            initialdir='Bilder',
            title="Bild speichern")
        if dateiname in ['',()]: return
        self.zeichenfenster.postscript(file=dateiname)

    def zeichne(self):
        "zeichnet vollstaendig neu"
        wieOft=self.mSchritte//self.alleWieviel
        _thread.start_new_thread(self.Action, (0.5, wieOft))

    def Action(self, sec, anzahl):
        """schrittweise wird ..."""
        zf=self.zeichenfenster
        zf.create_rectangle(0.0,0.0,self.breite,self.breite,fill=bg)
        for punkte in self.daten:
            zf.delete('punkte')
            for j in range(len(punkte)):
                x,y=punkte[j]
                zf.create_rectangle(x,y,x+2,y+2,outline=punktFarbe,tag='punkte')
            time.sleep(sec)
    
### ---------------------------------------------------
if __name__ == '__main__':
    print("Bitte warten!")
    print("Erst nach Ende der Berechnung wird gezeichnet.")
    root = Tk()
    app = grafikApp(master=root)
    app.master.title('Grafik')
    print('Bild speichern mit Doppelklick auf die Zeichenflaeche')
    app.mainloop()
    try:
        root.destroy()
    except Exception as e:
        pass


